    @echo off
    REM **************************************************************************
    REM * build_recovery_image_dismexit.cmd
    REM *
    REM * (c) Copyright by IBM Corporation and other(s) 2018, 2021. All Rights
    REM * Reserved.
    REM *
    REM * This script is called by build_recovery_image.cmd after a WinPE image
    REM * is mounted. For more information, see "READ ME CAREFULLY" below.
    REM *
    REM * This script is provided "as is", without warranty of any kind.
    REM *
    REM * Author: Andrew M. Raibeck
    REM *         IBM Spectrum Protect Development
    REM *         Level 3 Team Lead
    REM *         storman@us.ibm.com
    REM *
    REM * Version 2.3 - 2022-01-04 - Resequence the version history in most-to-
    REM *                            least recent order.
    REM *
    REM * Version 2.2 - 2021-08-30 - Change version number to 2.2 to align with
    REM *                            updated restore_system.cmd script.
    REM *
    REM * Version 2.1 - 2021-02-16 - No changes, code just adjust this version
    REM *                            number to match the accompanying
    REM *                            build_recovery_image script.
    REM *
    REM * Version 1.3 - 2019-07-22 - Bug fixes:
    REM *                               - Enable verbose logging if IBM_SP_DEBUG,
    REM *                                 set by the calling script, is "true".
    REM *                               - Added this history log section.
    REM *
    REM * Version 1.2 - 2018-11-13 - Initial version, numbered to match the
    REM *                            corresponding build_recovery_image.cmd
    REM *                            version.
    REM **************************************************************************

    if "%IBM_SP_DEBUG%" == "true" @echo on
    REM ---------------- PRE-FLIGHT CHECKS, DO NOT TOUCH!!! ----------------
    if "%IBM_SP_WINPE_DIR%" equ "" goto :WINPE_DIR_NOT_SET
    if not exist "%IBM_SP_WINPE_DIR%\mount" goto :WINPE_MOUNT_DIR_NOT_FOUND
    set IBM_SP_DISMEXIT_RC=0

    REM Skip past the example README and example customization
    goto :CUSTOMIZE


    REM -------------------- READ ME CAREFULLY --------------------
    REM This is a customizable exit called by the main "build_recovery_image.cmd"
    REM script. It is called while the Windows PE image is mounted for servicing.
    REM You can add your own DISM commands to inject additional packages or drivers
    REM between the comments below that say 'BEGIN user customization' and 'END user
    REM customization'.
    REM
    REM Tip: Make sure the build_recovery_image.cmd script can successfully create a
    REM WinPE image before tailoring this script. Doing so will ensure that the
    REM WinPE tools are working correctly.
    REM
    REM Make sure you specify "%IBM_SP_WINPE_DIR%\mount", with quotation marks, to
    REM refer to the directory where the image is mounted. The environment variable
    REM IBM_SP_WINPE_DIR is set by the build_recovery_image script. DO NOT CHANGE
    REM THIS ENVIRONMENT VARIABLE.
    REM
    REM After each command, be sure to test %ERRORLEVEL%, and issue a message if the
    REM command was unsuccessful. Set environment variable IBM_SP_DISMEXIT_RC to a
    REM non-zero return code, then go to :LEAVE_EXIT. This will cause the
    REM build_recovery_image.cmd script to stop.
    REM
    REM This script must reside in the same directory as build_recovery_image.cmd.
    REM
    REM If your custom commands are successful, then goto :LEAVE_EXIT.
    REM
    REM Things to avoid:
    REM
    REM    - Do not mount or dismount the servicing image, only use DISM commands to
    REM      add packages or drivers.
    REM
    REM    - Do not change the name of this script. It must be named
    REM      "build_recovery_image_dismexit.cmd" and it must reside in the same
    REM      directory as "build_recovery_image.cmd".
    REM
    REM    - Do not change the value of the IBM_SP_WINPE_DIR environment variable.
    REM
    REM Before adding your own customizations, see the example below, which
    REM illustrates best practices for robust behavior.



    REM -------------------- BEGIN example customization --------------------
    REM
    REM Example: Add VMware drivers to the Windows PE image. This example code
    REM assumes that you are running the Windows ADK on a system where VMware Tools
    REM is installed. This example was created on a Windows 10 workstation with
    REM VMware Tools 10.3.2 installed. You might want to add these drivers if your
    REM restore target is a VMware virtual machine.
    REM
    REM In this example, the pvscsi and vmxnet3 drivers are expected to exist, but
    REM the vmxnet driver might not exist. Pay attention to the logic that:
    REM
    REM    - validates assumptions (needed directories exist)
    REM    - tests the return code after each command finishes
    REM    - handles non-zero return codes (error conditions) from DISM by issuing
    REM      an appropriate error message, setting IBM_SP_DISMEXIT_RC to a non-zero
    REM      value, and using "goto :LEAVE_EXIT" to exit the script
    REM    - uses "goto :LEAVE_EXIT" to exit the script if the drivers are added
    REM      successfully.


    REM Validate assumptions: Verify that expected directories exist.
    if not exist "C:\Program Files\Common Files\VMware\Drivers\pvscsi" goto :NO_PVSCSI
    if not exist "C:\Program Files\Common Files\VMware\Drivers\vmxnet3" goto :NO_VMXNET3

    REM Add the drivers. A non-zero return code from DISM means that the
    REM command was unsuccessful.
    dism /image:"%IBM_SP_WINPE_DIR%\mount" /add-driver /driver:"C:\Program Files\Common Files\VMware\Drivers\pvscsi" /recurse
    if %ERRORLEVEL% neq 0 goto :DISM_ADD_PVSCSI_FAILED

    dism /image:"%IBM_SP_WINPE_DIR%\mount" /add-driver /driver:"C:\Program Files\Common Files\VMware\Drivers\vmxnet3" /recurse
    if %ERRORLEVEL% neq 0 goto :DISM_ADD_VMXNET3_FAILED

    REM Add optional driver if it exists (but fail if the attempt to add is
    REM unsuccessful).
    if not exist "C:\Program Files\Common Files\VMware\Drivers\vmxnet" goto :NO_VMXNET
    dism /image:"%IBM_SP_WINPE_DIR%\mount" /add-driver /driver:"C:\Program Files\Common Files\VMware\Drivers\vmxnet" /recurse
    if %ERRORLEVEL% neq 0 goto :DISM_ADD_VMXNET_FAILED
    goto :LEAVE_EXIT

    REM Error handling - display error messages, set IBM_SP_DISMEXIT_RC to a non-zero
    REM value, then go to :LEAVE_EXIT.
:NO_PVSCSI
    echo.
    echo -------------------------------------------------------------------------------
    echo - "C:\Program Files\Common Files\VMware\Drivers\pvscsi" was expected to exist
    echo - but could not be found.
    echo -------------------------------------------------------------------------------
    set IBM_SP_DISMEXIT_RC=1
    goto :LEAVE_EXIT

:DISM_ADD_PVSCSI_FAILED
    echo.
    echo -------------------------------------------------------------------------------
    echo - The pvscsi package could not be successfully added to the WinPE image.
    echo -------------------------------------------------------------------------------
    set IBM_SP_DISMEXIT_RC=1
    goto :LEAVE_EXIT

:NO_VMXNET3
    echo.
    echo -------------------------------------------------------------------------------
    echo - "C:\Program Files\Common Files\VMware\Drivers\vmxnet3" was expected to exist
    echo - but could not be found.
    echo -------------------------------------------------------------------------------
    set IBM_SP_DISMEXIT_RC=1
    goto :LEAVE_EXIT

:DISM_ADD_VMXNET3_FAILED
    echo.
    echo -------------------------------------------------------------------------------
    echo - The vmxnet3 package could not be successfully added to the WinPE image.
    echo -------------------------------------------------------------------------------
    set IBM_SP_DISMEXIT_RC=1
    goto :LEAVE_EXIT

:NO_VMXNET
    echo.
    echo -------------------------------------------------------------------------------
    echo - "C:\Program Files\Common Files\VMware\Drivers\vmxnet" was not found, so it
    echo - will not be added to the WinPE image.
    echo -------------------------------------------------------------------------------
    goto :LEAVE_EXIT

:DISM_ADD_VMXNET_FAILED
    echo.
    echo -------------------------------------------------------------------------------
    echo - The vmxnet package could not be successfully added to the WinPE image.
    echo -------------------------------------------------------------------------------
    set IBM_SP_DISMEXIT_RC=1
    goto :LEAVE_EXIT
REM -------------------- END example customization --------------------




:CUSTOMIZE
    REM -------------------- BEGIN user customization --------------------
    REM
    REM Best practice: verify that expected directories or files exist, and exit
    REM with a non-zero return code if they do not exist. In some cases, the
    REM directories might be optional, but it is up to the user to ensure the
    REM accuracy of the commands and to add the necessary error detection. Taking
    REM care to do this now will avoid unpleasant consequences later if the WinPE
    REM image does not work as expected. See the example above.
    REM
    REM --------------------  END user customization  --------------------



    REM -------------------- DO NOT TOUCH BELOW THIS LINE!!! --------------------
    GOTO :LEAVE_EXIT

:WINPE_DIR_NOT_SET
    echo.
    echo -------------------------------------------------------------------------------
    echo - Environment variable IBM_SP_WINPE_DIR is not set, perhaps this script was
    echo - not invoked by the build_recovery_image.cmd script.
    echo -------------------------------------------------------------------------------
    set IBM_SP_DISMEXIT_RC=12
    goto :LEAVE_EXIT

:WINPE_MOUNT_DIR_NOT_FOUND
    echo.
    echo -------------------------------------------------------------------------------
    echo - "%IBM_SP_WINPE_DIR\mount" could not be found.
    echo -------------------------------------------------------------------------------
    set IBM_SP_DISMEXIT_RC=12
    goto :LEAVE_EXIT

:LEAVE_EXIT
    exit /b %IBM_SP_DISMEXIT_RC%
