    @echo off
 
    set IBM_SP_SCRIPT_VERSION=2.3
    REM **************************************************************************
    REM * restore_system.cmd
    REM *
    REM * (c) Copyright by IBM Corporation and other(s) 2021. All Rights Reserved.
    REM *
    REM * This script guides the user through the IBM Spectrum Protect restore of
    REM * system state and local drives from within a WinPE environment.
    REM *
    REM * This script is a sample only, and is provided on an "as is" basis,
    REM * without warranty of any kind.
    REM *
    REM * Author: Andrew M. Raibeck
    REM *         IBM Spectrum Protect Development
    REM *         Level 3 Team Lead
    REM *         storman@us.ibm.com
    REM *
    REM * Version 2.3 - 2022-01-04 - Fix a problem with parsing QUERY FILESPACE
    REM *                            output when node replication is used. This
    REM *                            problem causes:
    REM *                              - EFI file restore fails
    REM *                              - The drive restore script is empty
    REM *                          - Resequence the version history in most-to-
    REM *                            least recent order.
    REM *                          - Reposition the IBM_SCRIPT_VERSION environment
    REM *                            variable setting to the top of the script (to
    REM *                            make it easier to remember to update it!).
    REM *                          - Fixed a bug where the script tried to create
    REM *                            a diagnostic log bundle when the script is
    REM *                            not running in WinPE. In this case, the log
    REM *                            bundle should not be created.
    REM *
    REM * Version 2.2 - 2021-09-15 - Add support for point-in-time (PIT) restore.
    REM *                          - Provide option to open a command prompt to
    REM *                            make custom manual network setting changes.
    REM *                            Copy all restore "bootstrap" files into the
    REM *                            X:\ibm_sp_recovery directory. Create a zip
    REM *                            file of X:\ibm_sp_recovery for diagnostic
    REM *                            purposes.
    REM *                          - Capture VSS_ASR trace in system state restore
    REM *                            for diagnostic purposes.
    REM *                          - Include a tip to the drive restore script to
    REM *                            add the "-quiet" option for a drive restore
    REM *                            that has millions of files. This will prevent
    REM *                            the X: drive from filling up due to the size
    REM *                            of the redirected dsmc output file.
    REM *
    REM * Version 2.1 - 2021-02-16 - Add step numbers. Up-left version to match
    REM *                            the accompanying build_recovery_image script.
    REM *
    REM * Version 1.1 - 2021-01-14 - Initial version
    REM **************************************************************************

    REM This script needs to run on the system drive. If run from the WinPE boot
    REM media, copy it to the system drive, then re-start the script. In this
    REM context, "system drive" refers to the drive pointed to by the
    REM %SystemDrive% environment variable (usually X:).
    if /i "%~d0" neq "%SystemDrive%" (
        copy /y "%~d0%~p0%~nx0" "%SystemDrive%\%~nx0"
        cd /d % %SystemDrive%\
        "%SystemDrive%\%~nx0"
    )

    REM Later, this directory is assigned to environment variable IBM_SP_RECOVERY_DIR.
    if not exist "%SystemDrive%\ibm_sp_recovery" md "%SystemDrive%\ibm_sp_recovery"
    if not exist "%SystemDrive%\ibm_sp_recovery" goto :CANNOT_CREATE_WORKING_DIR
    cd /d "%SystemDrive%\ibm_sp_recovery"

    REM make the screen height long enough to prevent output from scrolling off.
    mode con: lines=9999

    setlocal enabledelayedexpansion

    echo ATTENTION: This command prompt window's buffer is increased to retain more
    echo message history. Before proceeding, take the time to reposition this window
    echo so it is fully visible on the screen.
    echo.
    set /p IBM_SP_USER_RESP=Press 'ENTER' when you are ready to continue =^> 
    echo.

    set IBM_SP_DEBUG=false
    if "%1" equ "-debug" (
        set IBM_SP_DEBUG=true
        echo on
    )

    REM %0 is the full path used to invoke the script.
    REM %~nx0 gets just the file name and extension of the script.
    set IBM_SP_SCRIPT_NAME=%~nx0

    set IBM_SP_NAME=IBM Spectrum Protect
    set IBM_SP_RECOVERY_DIR=%cd%
    set IBM_SP_RECOVERY_LOG=%IBM_SP_RECOVERY_DIR%\recovery.log
    set IBM_SP_ZIP_BUNDLE=%IBM_SP_RECOVERY_DIR%_log_bundle.zip
    set IBM_SP_RESTORE_SYSTEM_STATE_LOG=%IBM_SP_RECOVERY_DIR%\dsmc_restore_systemstate.log
    set IBM_SP_DSM_OPT=%IBM_SP_RECOVERY_DIR%\dsm.restore.opt
    set IBM_SP_DUMMY_ID_FILE=_____I_AM_WINPE_BOOT_MEDIA_WITH_IBM_SPECTRUM_PROTECT_CLIENT______
    set IBM_SP_STRIPPED1_LOG=dsmerror_stripped1.log
    set IBM_SP_STRIPPED2_LOG=dsmerror_stripped2.log

    REM Create the little helper script that sleeps for the specified number of seconds.
    set IBM_SP_WAIT=%IBM_SP_RECOVERY_DIR%\ibm_sp_sleep.vbs
    echo wait_time = 10     ' seconds>"%IBM_SP_WAIT%"
    echo If WScript.Arguments.Count = 1 Then>>"%IBM_SP_WAIT%"
    echo wait_time = WScript.Arguments.Item(0)>>"%IBM_SP_WAIT%"
    echo End If>>"%IBM_SP_WAIT%"
    echo WScript.Sleep(wait_time * 1000)>>"%IBM_SP_WAIT%"

    echo.
    echo -------------------------------------------------------------------------------
    echo - %IBM_SP_SCRIPT_NAME% Version %IBM_SP_SCRIPT_VERSION%
    echo -
    echo - (c) Copyright by IBM Corporation and other(s) 2021. All Rights Reserved.
    echo -
    echo - This script guides you through the IBM Spectrum Protect restore of system
    echo - state and local drives from within a WinPE environment.
    echo -------------------------------------------------------------------------------
    echo.

    call :DISPLAY_STEP_NUMBER
    echo -------------------------------------------------------------------------------
    echo - Verifying that the system is running in WinPE
    echo -------------------------------------------------------------------------------
    set IBM_SP_REGKEY_MININT=HKLM\SYSTEM\ControlSet001\Control\MiniNT
    reg query %IBM_SP_REGKEY_MININT% /ve >nul 2>&1
    if %ERRORLEVEL% neq 0 goto :NOT_RUNNING_IN_WINPE_MODE
    echo The system appears to be running in WinPE.
    echo.

    REM Try to determine whether the system was booted with BIOS or UEFI.
    call :DISPLAY_STEP_NUMBER
    echo -------------------------------------------------------------------------------
    echo - Checking the firmware type for BIOS or UEFI
    echo -------------------------------------------------------------------------------
    wpeutil UpdateBootInfo >nul 2>&1
    set IBM_SP_REGKEY_CONTROL=HKLM\SYSTEM\CurrentControlSet\Control
    for /f "tokens=3" %%a in ('reg query %IBM_SP_REGKEY_CONTROL% /v PEFirmwareType') do set IBM_SP_REGDATA=%%a
    if "%IBM_SP_REGDATA%" equ "0x1" (
        set IBM_SP_FIRMWARE=BIOS
        goto :FIRMWARE_DETECTED
    ) else if "%IBM_SP_REGDATA%" equ "0x2" (
        set IBM_SP_FIRMWARE=UEFI
        goto :FIRMWARE_DETECTED
    ) else (
        echo The script could not determine whether the system was booted with BIOS or UEFI.
        goto :CHOOSE_FIRMWARE_TYPE
    )


:FIRMWARE_DETECTED
    echo The script detected that this system was booted with %IBM_SP_FIRMWARE% firmware.


:FIRMWARE_YN_PROMPT
    REM The "set /p" commands have a blank space at the end to separate the user input from the prompt!
    set /p IBM_SP_USER_RESP=Is %IBM_SP_FIRMWARE% the correct firmware for the system you want to restore (y/n/q) ? 
    if /i "%IBM_SP_USER_RESP%" equ "y" goto :PRE_DISKPART
    if /i "%IBM_SP_USER_RESP%" equ "yes" goto :PRE_DISKPART
    if /i "%IBM_SP_USER_RESP%" equ "n" goto :SWITCH_FIRMWARE_TYPE
    if /i "%IBM_SP_USER_RESP%" equ "no" goto :SWTICH_FIRMWARE_TYPE
    if /i "%IBM_SP_USER_RESP%" equ "q" goto :FIRMWARE_UNKNOWN
    if /i "%IBM_SP_USER_RESP%" equ "quit" goto :FIRMWARE_UNKNOWN
    echo.
    echo Enter 'y' for 'yes' or 'n' for 'no'. If you are unsure, enter 'q' to 'quit'.
    goto :FIRMWARE_YN_PROMPT


:SWITCH_FIRMWARE_TYPE
    if "%IBM_SP_FIRMWARE%" equ "BIOS" (
        set IBM_SP_FIRMWARE=UEFI
    ) else (
        set IBM_SP_FIRMWARE=BIOS
    )
    goto :FIRMWARE_YN_PROMPT


:CHOOSE_FIRMWARE_TYPE
    set /p IBM_SP_USER_RESP=Enter 'b' if the firmware is BIOS or 'u' if the firmware is UEFI (b/u/q) =^> 
    if /i "%IBM_SP_USER_RESP%" equ "b" goto :SET_BIOS
    if /i "%IBM_SP_USER_RESP%" equ "bios" goto :SET_BIOS
    if /i "%IBM_SP_USER_RESP%" equ "u" goto :SET_UEFI
    if /i "%IBM_SP_USER_RESP%" equ "uefi" goto :SET_UEFI
    if /i "%IBM_SP_USER_RESP%" equ "q" goto :FIRMWARE_UNKNOWN
    if /i "%IBM_SP_USER_RESP%" equ "quit" goto :FIRMWARE_UNKNOWN
    echo.
    echo Enter 'b' for 'BIOS' or 'u' for 'UEFI'. If you are unsure, enter 'q' to quit.
    goto :CHOOSE_FIRMWARE_TYPE


:SET_BIOS
    set IBM_SP_FIRMWARE=BIOS
    goto :FIRMWARE_YN_PROMPT


:SET_UEFI
    set IBM_SP_FIRMWARE=UEFI
    goto :FIRMWARE_YN_PROMPT


:PRE_DISKPART
    call :DISPLAY_STEP_NUMBER
    echo -------------------------------------------------------------------------------
    echo - Verify that disks are clean (not partitioned, not formatted)
    echo -------------------------------------------------------------------------------
    echo This restore method works best when restoring to unpartitioned, unformatted
    echo disks. Otherwise the restore might not succeed.
    echo.
    echo If the disks have any data or partitions on them, use DISKPART to select and
    echo clean each disk. After the disks are cleaned, WinPE is rebooted.
:PRE_DISKPART1
    echo.
    set /p IBM_SP_USER_RESP=Do you want to run DISKPART (y/n/q) ? 
    if /i "%IBM_SP_USER_RESP%" equ "y" goto :LAUNCH_DISKPART
    if /i "%IBM_SP_USER_RESP%" equ "yes" goto :LAUNCH_DISKPART
    if /i "%IBM_SP_USER_RESP%" equ "n" goto :CREATE_OPTIONS_FILE
    if /i "%IBM_SP_USER_RESP%" equ "no" goto :CREATE_OPTIONS_FILE
    if /i "%IBM_SP_USER_RESP%" equ "q" goto :USER_QUIT
    if /i "%IBM_SP_USER_RESP%" equ "quit" goto :USER_QUIT
    echo.
    echo Enter 'y' to run DISKPART, 'n' to skip DISKPART, or 'q' to quit.
    goto :PRE_DISKPART1


:LAUNCH_DISKPART
    echo.
    echo Use command LIST DISK to list all disks.
    echo Use command SELECT DISK n to select disk number n.
    echo Use command CLEAN to remove partitions and formatting from the selected disk.
    echo Repeat the SELECT DISK and CLEAN commands for each disk shown by LIST DISK.
    echo Use command EXIT when you are done.
    echo.
    echo For example, if LIST DISK shows DISK 0, DISK 1, and DISK 2, then enter this
    echo sequence of commands to clean each disk:
    echo.
    echo     select disk 0
    echo     clean
    echo     select disk 1
    echo     clean
    echo     select disk 2
    echo     clean
    echo     exit
    echo.
    echo After the disks are cleaned, reboot WinPE, then restart this script. Then you
    echo can skip the DISKPART step and proceed with the rest of the restore operation.
    echo.
    echo ***** IMPORTANT *****
    echo If you booted from a USB flash drive, do not clean the flash drive disk.
    echo.
    set /p IBM_SP_USER_RESP=Press 'ENTER' to start DISKPART =^> 
    echo.
    start /wait diskpart


:POST_DISKPART
    echo.
    echo If you cleaned any disks, the system must be restarted. If no disks needed
    echo cleaning, you can continue with the next step. If necessary, you can run
    echo DISKPART again.
:POST_DISKPART1
    echo.
    set /p IBM_SP_USER_RESP=Enter 'r' to restart WinPE, 'c' to continue, 'd' to run DISKPART (r/c/d/q) =^> 
    if /i "%IBM_SP_USER_RESP%" equ "r" wpeutil reboot
    if /i "%IBM_SP_USER_RESP%" equ "restart" wpeutil reboot
    if /i "%IBM_SP_USER_RESP%" equ "c" goto :CREATE_OPTIONS_FILE
    if /i "%IBM_SP_USER_RESP%" equ "continue" goto :CREATE_OPTIONS_FILE
    if /i "%IBM_SP_USER_RESP%" equ "d" goto :LAUNCH_DISKPART
    if /i "%IBM_SP_USER_RESP%" equ "diskpart" goto :LAUNCH_DISKPART
    if /i "%IBM_SP_USER_RESP%" equ "q" goto :USER_QUIT
    if /i "%IBM_SP_USER_RESP%" equ "quit" goto :USER_QUIT
    goto :POST_DISKPART1


:CREATE_OPTIONS_FILE
    call :DISPLAY_STEP_NUMBER
    echo -------------------------------------------------------------------------------
    echo - Create the backup-archive client options file
    echo -------------------------------------------------------------------------------
    echo You will be prompted to enter information needed to create the options file
    echo used to connect to the %IBM_SP_NAME% server. You can modify the options
    echo file later.
    echo.
    echo Note: Before proceeding further, your %IBM_SP_NAME% server administrator
    echo might need to update the node's SESSIONSECURITY attribute to TRANSITIONAL.
    echo.

    set /p IBM_SP_TCPS=Enter the TCP/IP address of the %IBM_SP_NAME% server =^> 
    echo.
    set /p IBM_SP_TCPP=Enter the TCP/IP port number of the %IBM_SP_NAME% server (default is 1500) =^> 
    if "%IBM_SP_TCPP%" == "" set %IBM_SP_TCPP=1500
    echo.
    set /p IBM_SP_NODE=Enter the node name =^> 
    echo.

    REM Create a basic options file. Note that the '^' character is an escape
    REM character.
    echo * Review the following settings for accuracy. You can also adjust any values as> "%IBM_SP_DSM_OPT%"
    echo * needed. When you are finished, save any changes, then close Notepad to>> "%IBM_SP_DSM_OPT%"
    echo * continue.>> "%IBM_SP_DSM_OPT%"
    echo.>> "%IBM_SP_DSM_OPT%"
    echo tcpserveraddress %IBM_SP_TCPS%>> "%IBM_SP_DSM_OPT%"
    echo tcpport %IBM_SP_TCPP%>> "%IBM_SP_DSM_OPT%"
    echo * TCPWINDOWSIZE 0 is often a good choice to get the system-tuned size.>> "%IBM_SP_DSM_OPT%"
    echo tcpwindowsize ^0>> "%IBM_SP_DSM_OPT%"
    echo.>> "%IBM_SP_DSM_OPT%"
    echo nodename %IBM_SP_NODE%>> "%IBM_SP_DSM_OPT%"
    echo passwordaccess generate>> "%IBM_SP_DSM_OPT%"
    echo.>> "%IBM_SP_DSM_OPT%"
    echo * The following setting displays the client return code after an operation>> "%IBM_SP_DSM_OPT%"
    echo * finishes.>> "%IBM_SP_DSM_OPT%"
    echo testflags showallconsistentrc>> "%IBM_SP_DSM_OPT%"

    REM Display the options file to the user for review.
    notepad "%IBM_SP_DSM_OPT%"


:PING_PROTECT_SERVER
    REM Check if the server is ping-able.
    call :DISPLAY_STEP_NUMBER
:PING_PROTECT_SERVER_2
    echo -------------------------------------------------------------------------------
    echo - Ping test to %IBM_SP_NAME% server to check basic network connectivity
    echo -------------------------------------------------------------------------------
    echo Attempting to ping %IBM_SP_TCPS% to check basic network connectivity.
    echo This might take a few moments ...
    ping %IBM_SP_TCPS% | find "TTL=" >nul 2>&1
    if %ERRORLEVEL% equ 0 (
        echo.
        echo The ping was successful.
        goto :SET_ENVIRONMENT_VARIABLES
    )

    REM The ping failed, so try using the netsh int command to set up the
    REM network interface.
    echo -------------------------------------------------------------------------------
    echo.
    echo The server cannot be pinged. The restore cannot continue if the server
    echo cannot be reached. Some reasons include the WinPE network must be manually
    echo configured, or a firewall prevents the server from responding to pings.
    echo.
    echo This script can try the 'netsh int' command to configure a static IP address
    echo for this WinPE instance. You need to know the following values:
    echo.
    echo     - Network interface name (example: Ethernet0)
    echo     - IP address for this WinPE instance
    echo     - Subnet mask (example: 255.255.255.0)
    echo     - Network gateway
    echo.
    echo If you do not know these values, contact your network administrator.
    echo.
    echo If you entered the incorrect IP address, you can repeat the steps to create the
    echo backup-archive client options file.
    echo.
    echo If the server does not respond to pings but the network is configured
    echo correctly, you can try to continue.
    echo.
    echo If you need to make other network configuration changes, you can open a command
    echo prompt from which you can make the changes.
    echo.


:NETSH_YN_PROMPT
    echo Enter 'r' to re-do the options file, 'i' to configure a static IP address,
    set /p IBM_SP_USER_RESP=      'p' for a command prompt, 'c' to continue, or 'q' quit (r/i/p/c/q) =^> 
    if /i "%IBM_SP_USER_RESP%" equ "r" goto :CREATE_OPTIONS_FILE
    if /i "%IBM_SP_USER_RESP%" equ "i" goto :GET_STATIC_IP_CONFIG_INFO
    if /i "%IBM_SP_USER_RESP%" equ "p" goto :DO_MANUAL_CONFIG
    if /i "%IBM_SP_USER_RESP%" equ "c" goto :SET_ENVIRONMENT_VARIABLES
    if /i "%IBM_SP_USER_RESP%" equ "q" goto :CANNOT_CONNECT_TO_PROTECT_SERVER
    echo.
    goto :NETSH_YN_PROMPT


:GET_STATIC_IP_CONFIG_INFO
    echo -------------------------------------------------------------------------------
    echo You will be prompted to enter information needed to configure the static IP
    echo address. If you make a mistake, such as a typographical error, do not worry:
    echo you can re-enter the information.
    echo.
    echo Available network interfaces:
    netsh interface show interface
    set /p IBM_SP_NETWORK_INTERFACE=Enter the name of the network interface =^> 
    echo.
    set /p IBM_SP_IP_ADDRESS=Enter the static IP address for this WinPE instance =^> 
    echo.
    set /p IBM_SP_SUBNET_MASK=Enter the subnet mask =^> 
    echo.
    set /p IBM_SP_GATEWAY_ADDRESS=Enter the address of the network gateway =^> 
    echo.

    REM Let the user review the answers
    echo You entered the following information:
    echo.
    echo     - Network interface name: %IBM_SP_NETWORK_INTERFACE%
    echo     - Static IP address: %IBM_SP_IP_ADDRESS%
    echo     - Subnet mask: %IBM_SP_SUBNET_MASK%
    echo     - Network gateway: %IBM_SP_GATEWAY_ADDRESS%
    echo.


:STATIC_INFO_CONFIRMATION
    set /p IBM_SP_USER_RESP=Is this correct (y/n) ? 
    if /i "%IBM_SP_USER_RESP%" equ "y" goto :CONFIGURE_STATIC_IP_ADDRESS
    if /i "%IBM_SP_USER_RESP%" equ "yes" goto :CONFIGURE_STATIC_IP_ADDRESS
    if /i "%IBM_SP_USER_RESP%" equ "n" goto :CANNOT_CONNECT_TO_PROTECT_SERVER
    if /i "%IBM_SP_USER_RESP%" equ "no" goto :CANNOT_CONNECT_TO_PROTECT_SERVER
    echo.
    echo Enter 'y' for 'yes' or 'n' for 'no'.
    goto :STATIC_INFO_CONFIRMATION


:CONFIGURE_STATIC_IP_ADDRESS
    REM Command is: netsh int ip set address "<interface_name>" static <ip_address> <subnet_mask> <gateway>
    set IBM_SP_NETSH_CMD=netsh int ip set address "%IBM_SP_NETWORK_INTERFACE%" static %IBM_SP_IP_ADDRESS% %IBM_SP_SUBNET_MASK% %IBM_SP_GATEWAY_ADDRESS%
    echo.
    echo Running the following command to configure the static IP address:
    echo %IBM_SP_NETSH_CMD%
    %IBM_SP_NETSH_CMD%
    REM In my test lab, it sometimes took a few seconds for the configuration to
    REM go into effect. Therefore, I added an (admittedly abitrary) wait of 10
    REM seconds before trying the ping again.
    cscript "%IBM_SP_WAIT%" 10 > nul 2>&1
    goto :PING_PROTECT_SERVER_2


:DO_MANUAL_CONFIG
    echo.
    echo A new command prompt from which you can configure the network will be opened.
    echo When you are done, type 'EXIT' from within the new command prompt to return to
    echo this script.
    echo.
    set /p IBM_SP_USER_RESP=Press 'ENTER' to start the new command prompt. =^> 
    start /wait cmd.exe
    goto :PING_PROTECT_SERVER_2


:SET_ENVIRONMENT_VARIABLES
    set DSM_LOG=%IBM_SP_RECOVERY_DIR%
    set DSM_CONFIG=%IBM_SP_DSM_OPT%
    set IBM_SP_ORIGINAL_PATH=%PATH%
    call :FIND_CD_DRIVE
    if "%IBM_SP_CLIENT_DRIVE%" equ "" goto :CANNOT_FIND_CD_DRIVE
    call :SET_OTHER_ENVIRONMENT_VARIABLES


:DO_QUERY_SESSION
    call :DISPLAY_STEP_NUMBER
    echo -------------------------------------------------------------------------------
    echo - Testing backup-archive client connectivity to the %IBM_SP_NAME%
    echo - server.
    echo -
    echo - Respond to any prompts for user ID and password.
    echo -------------------------------------------------------------------------------
    dsmc query session
    if %ERRORLEVEL% equ 0 (
        echo.
        echo The client successfully connected to the server.
        goto :ASK_IF_THIS_IS_A_PIT_RESTORE
    )

    echo.
    echo The backup-archive client could not connect to the %IBM_SP_NAME% server.


:QUERY_SESSION_ERROR_HANDLER
    echo.
    set /p IBM_SP_USER_RESP=Enter 't' to try connecting again or 'c' to check the environment (t/c/q) =^> 
    if /i "%IBM_SP_USER_RESP%" equ "t" goto :DO_QUERY_SESSION
    if /i "%IBM_SP_USER_RESP%" equ "c" goto :QUERY_SESSION_CHECK_ENVIRONMENT
    if /i "%IBM_SP_USER_RESP%" equ "q" goto :CANNOT_QUERY_SESSION
    echo.
    echo Enter 't' to try the QUERY SESSION operation again or 'c' to manually check the
    echo environment. If you want to quit, enter 'q'.
    goto :QUERY_SESSION_ERROR_HANDLER


:QUERY_SESSION_CHECK_ENVIRONMENT
    echo.
    echo A new command prompt from which you can check the environment will be opened.
    echo When you are done, type 'EXIT' from within the new command prompt to return
    echo to this script.
    echo.
    set /p IBM_SP_USER_RESP=Press 'ENTER' to start the new command prompt. =^> 
    start /wait cmd.exe
    goto :QUERY_SESSION_ERROR_HANDLER


:ASK_IF_THIS_IS_A_PIT_RESTORE
    call :DISPLAY_STEP_NUMBER
    echo -------------------------------------------------------------------------------
    echo - Checking whether this is a point-in-time (PIT) restore.
    echo -------------------------------------------------------------------------------
    REM ATTENTION: Do not remove the blank space at the end of the following line.
    REM Blank space here --v
    set IBM_SP_PIT_OPTIONS= 
    REM Blank space here --^
:ASK_IF_THIS_IS_A_PIT_RESTORE_2
    echo.
    set /p IBM_SP_USER_RESP=Is this a point-in-time restore (y/n) ? 
    if /i "%IBM_SP_USER_RESP%" equ "n" goto :DO_RESTORE_SYSTEM_STATE
    if /i "%IBM_SP_USER_RESP%" equ "y" goto :GET_PIT_CRITERIA
    echo.
    echo Enter 'y' for 'yes' or 'n' for 'no'.
    goto :ASK_IF_THIS_IS_A_PIT_RESTORE_2


:GET_PIT_CRITERIA
    echo.
    echo When prompted, specify a date and time that represents the date and time of the
    echo backup version you want to restore.
    echo.
    echo Attention: The date and time must be the same or later than the date and time
    echo the entire backup (system state and drives) ended, but before the next backup
    echo started.
:GET_PIT_CRITERIA_2
    echo.
    echo Enter the date in mm/dd/yyyy format, then enter the time in HH:mm:ss format.
    echo.
    echo For example, if the desired point in time is 9:50:00 PM on 7 August 2021, then
    echo enter 08/07/2021 for the date, and 21:50:00 for the time.
    echo.
    set /p IBM_SP_PIT_DATE=Enter the point-in-time restore date (mm/hh/yyyy) =^> 
    echo.
    set /p IBM_SP_PIT_TIME=Enter the point-in-time restore date (HH:mm:ss) =^> 
    echo.

    REM Let the user review the answers
    echo You entered the following information:
    echo.
    echo     - Point-in-time restore date: %IBM_SP_PIT_DATE%
    echo     - Point-in-time restore time: %IBM_SP_PIT_TIME%
    echo.


:PIT_CRITERIA_CONFIRMATION
    set /p IBM_SP_USER_RESP=Is this correct (y/n) ? 
    if /i "%IBM_SP_USER_RESP%" equ "y" goto :SET_PIT_CRITERIA
    if /i "%IBM_SP_USER_RESP%" equ "yes" goto :SET_PIT_CRITERIA
    if /i "%IBM_SP_USER_RESP%" equ "n" goto :GET_PIT_CRITERIA_2
    if /i "%IBM_SP_USER_RESP%" equ "no" goto :GET_PIT_CRITERIA_2
    echo.
    echo Enter 'y' for 'yes' or 'n' for 'no'.
    goto :PIT_CRITERIA_CONFIRMATION


:SET_PIT_CRITERIA
    set IBM_SP_PIT_OPTIONS=-dateformat=1 -timeformat=1 -pitdate=%IBM_SP_PIT_DATE% -pittime=%IBM_SP_PIT_TIME%


:DO_RESTORE_SYSTEM_STATE
    REM       First, start another script that runs concurrently to check for restore
    REM       of the tsmvolmap.txt file, then copy that file to the root of X:
    set IBM_SP_COPY_TSMVOLMAP_SCRIPT=%IBM_SP_RECOVERY_DIR%\ibm_sp_copy_tsmvolmap_file.cmd
    set IBM_SP_TSMVOLMAP_FILE=%IBM_SP_RECOVERY_DIR%\ibm_sp_tsmvolmap.txt
    REM The wait time is expressed in seconds
    set IBM_SP_WAIT_TIMER=10

    if exist "%temp%\~adsm.sys" rd /s /q "%temp%\~adsm.sys"

    REM Get rid of any residual copy of the tsmvolmap.txt file (but unless this
    REM script is restarted, it is not expected that this file exists).
    if exist "%IBM_SP_TSMVOLMAP_FILE%" delete "%IBM_SP_TSMVOLMAP_FILE%"

    REM Create the script that waits for the tsmvolmap.txt file to be restored,
    REM then copies it to the recovery directory. This script also copies the
    REM other restore "bootstrap" files to the recovery directory.
    echo     @echo off>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo     setlocal>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo.>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo :TOP>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo     dir "%%temp%%\~adsm.sys\tsmvolmap.txt" /s /b ^>nul 2^>^&^1>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo     if %%errorlevel%% equ 0 goto :COPY_FILE>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo     echo tsmvolmap.txt is not restored yet, waiting for %IBM_SP_WAIT_TIMER% seconds...>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo     cscript "%IBM_SP_WAIT%" %IBM_SP_WAIT_TIMER%>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo     goto :TOP>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo.>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo :COPY_FILE>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo     echo tsmvolmap.txt is restored, copying it to "%IBM_SP_TSMVOLMAP_FILE%">>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo     for /f "tokens=*" %%%%a in ('dir "%%temp%%\~adsm.sys\tsmvolmap.txt" /s /b') do copy "%%%%a" "%IBM_SP_TSMVOLMAP_FILE%">>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo     xcopy /e /h "%%temp%%\~adsm.sys\" "%IBM_SP_RECOVERY_DIR%\adsm.sys\">>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo.>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo :DONE>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo     endlocal>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"
    echo     exit>>"%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"

    REM Start the script
    call :DISPLAY_STEP_NUMBER
    echo -------------------------------------------------------------------------------
    echo - The system state restore operation is about to start. At the same time,
    echo - another script will be started. The second script waits for tsmvolmap.txt
    echo - to be restored, then copies the restored file to
    echo - "%IBM_SP_TSMVOLMAP_FILE%", and then ends.
    echo -
    echo - If something goes wrong with the system state restore, you can stop the
    echo - second script. Otherwise allow the script to finish.
    echo -
    echo - Another command prompt will be opened, from which you can examine IBM
    echo - Spectrum Protect output files while the restore operations run. Look for the
    echo - command prompt window titled "Secondary Command Prompt".
    echo -
    echo - The secondary script and the command prompt windows are initially minimized.
    echo -------------------------------------------------------------------------------
    echo.
    set /p IBM_SP_USER_RESP=Press 'ENTER' to start the system state restore =^> 
    echo.
    echo Waiting for system state restore to finish. The output is written to
    echo "%IBM_SP_RESTORE_SYSTEM_STATE_LOG%". This can take a while...

    start "Wait for tsmvolmap.txt restore" /min "%IBM_SP_COPY_TSMVOLMAP_SCRIPT%"

    REM Start the secondary prompt in the recovery directory.
    start "Secondary Command Prompt" /min cmd.exe
    popd

    REM Start the system state restore
    dsmc restore systemstate %IBM_SP_PIT_OPTIONS% -traceflags=pid,tid,vss_asr -tracefile=dsmc_restore_systemstate_asr_trace.log > "%IBM_SP_RESTORE_SYSTEM_STATE_LOG%" 2>&1
    set IBM_SP_ERRORLEVEL=%ERRORLEVEL%
    echo.
    echo Restore of the system state finished with return code %IBM_SP_ERRORLEVEL%
    REM Expected RCs are 0, 4, 8, or 12. 0 is good, 4 and 8 are probably okay,
    REM but 12 is a failure.
    if %IBM_SP_ERRORLEVEL% leq 8 echo The restore is considered successful.
    echo.
    if %IBM_SP_ERRORLEVEL% leq 8 goto :QUERY_FILE_SPACES

    REM Prepare a stripped version of the error log for user review. The
    REM following code removes messages that can be safely ignored and empty
    REM lines.

    REM Step 1. Get rid of empty lines.
    if exist "%IBM_SP_STRIPPED1_LOG%" del "%IBM_SP_STRIPPED1_LOG%"
    for /f "tokens=*" %%a in (dsmerror.log) do @echo %%a>>"%IBM_SP_STRIPPED1_LOG%"

    REM Step 2. Get rid of messages that can be safely ignored.
    echo ----- Top of stripped error log messages -----> "%IBM_SP_STRIPPED2_LOG%"
    type "%IBM_SP_STRIPPED1_LOG%" | find /v "ANS4042E" | find /v "ANS0361I" | find /v "ANS1376E" | find /v "ANS1997W" >> "%IBM_SP_STRIPPED2_LOG%"
    echo ----- Bottom of stripped error log messages ----->> "%IBM_SP_STRIPPED2_LOG%"

    echo The system state restore might still be successful. Review
    echo "%IBM_SP_STRIPPED2_LOG%". If there are any error messages, the restore might
    echo not be able to continue. If the file is empty, the restore can continue. If
    echo there are other error messages, then based on the errors, try to decide whether
    echo the restore can continue.
    echo.
    echo You can also use the other command prompt to open dsmerror.log and
    echo dsmc_restore_systemstate.log in Notepad to check for problems.
    echo.
    set /p IBM_SP_USER_RESP=Press 'ENTER' to review "%IBM_SP_STRIPPED2_LOG%" =^> 
    echo.

    REM Display the log file for the user to review.
    notepad "%IBM_SP_STRIPPED2_LOG%"

:ASK_TO_CONTINUE_RESTORE
    echo.
    set /p IBM_SP_USER_RESP=Do you want the restore to continue (y/n)? =^> 
    if /i "%IBM_SP_USER_RESP%" equ "y" goto :QUERY_FILE_SPACES
    if /i "%IBM_SP_USER_RESP%" equ "yes" goto :QUERY_FILE_SPACES
    if /i "%IBM_SP_USER_RESP%" equ "n" goto :SYSTEM_STATE_RESTORE_FAILED
    if /i "%IBM_SP_USER_RESP%" equ "no" goto :SYSTEM_STATE_RESTORE_FAILED
    echo.
    echo Enter 'y' for 'yes' or 'n' for 'no'.
    goto :ASK_TO_CONTINUE_RESTORE

:QUERY_FILE_SPACES
    REM Find local drive file spaces.
    set IBM_SP_FILE_SPACES=%IBM_SP_RECOVERY_DIR%\ibm_sp_dsmc_query_filespace.txt
    dsmc query filespace > "%IBM_SP_FILE_SPACES%" 2>&1

    if "%IBM_SP_FIRMWARE%" neq "UEFI" goto :DO_MOUNTVOL

    REM For UEFI firmware, restore the EFI files to the EFI partition.
    REM 1. Find system state file space name. Upon return from the call below,
    REM IBM_SP_SYSSTATE_FILESPACE is set to the file space name.
    call :SET_SYSTEM_STATE_FILE_SPACE_NAME

    REM It is very unexpected that the environment variable is not set, but it
    REM is always better to be paranoid.
    if "%IBM_SP_SYSSTATE_FILESPACE%" equ "" (
        echo Could not find the system state file space name in file
        echo "%IBM_SP_FILE_SPACES%". The restore will try to continue with
        echo restoring the drives. After the restore is finished, try to
        echo manually restore the EFI files by running these commands:
        echo.
        echo   mountvol q: /s
        echo.
        echo   dsmc restore "{xxxxxxx\SystemState\NULL\System State\SystemState}\EFI\Microsoft\*" q:\EFI\ -subdir=yes -replace=no %IBM_SP_PIT_OPTIONS% -optfile=%DSM_CONFIG% ^> %IBM_SP_RECOVERY_DIR%\dsmc_restore_efi_files.log 2^>^&1
        echo.
        echo   mountvol q: /d
        echo.
        echo Drive q: can be any unused drive.
        echo.
        echo Inspect the output of %IBM_SP_RECOVERY_DIR%\dsmc_restore_efi_files.log
        echo and very the restore completed without errors. Skipped files are okay.

        set /p IBM_SP_USER_RESP=Press 'ENTER' to continue the restore =^> 
        goto :DO_MOUNTVOL
    )

    REM 2. Find an unused drive letter where the EFI system partition can be
    REM temporarily mounted.
    set IBM_SP_EFI_DRIVE=
    for %%d in (A B C D E F G H I J K L M N O P Q R S T U V W X Y Z) do (
        if not exist "%%d:" set IBM_SP_EFI_DRIVE=%%d:
    )

    REM If no drive can be found, skip this step. Toward the end, a message
    REM is displayed to indicate the problem.
    if "%IBM_SP_EFI_DRIVE%" equ "" goto :DO_MOUNTVOL

    REM 3. Temporarily mount the EFI system partition.
    mountvol %IBM_SP_EFI_DRIVE% /s
    echo The EFI system partition is temporarily mounted to drive %IBM_SP_EFI_DRIVE%.

    REM 4. To avoid problems related to IT25446 or the case where the EFI data is
    REM restored to the wrong drive, restore the files to the EFI system partition,
    REM skipping any existing files.
    echo Starting restore of the EFI files. This will take a few moments...
    dsmc restore "{%IBM_SP_SYSSTATE_FILESPACE%}\EFI\Microsoft\*" %IBM_SP_EFI_DRIVE%\EFI\ -subdir=yes -replace=no %IBM_SP_PIT_OPTIONS% > %IBM_SP_RECOVERY_DIR%\dsmc_restore_efi_files.log 2>&1
    echo Restore of the EFI system partition files finished with return code %errorlevel%
    mountvol %IBM_SP_EFI_DRIVE% /d
    echo The EFI system partition is dismounted from drive %IBM_SP_EFI_DRIVE%.


:DO_MOUNTVOL
    call :DISPLAY_STEP_NUMBER
    echo -------------------------------------------------------------------------------
    echo - The system state restore finished. The next step is to mount the drives, then
    echo - restore them.
    echo -
    echo - In preparation for the restore, the target drive letters are dismounted, and
    echo - then remounted, to ensure the correct volume is mounted on the correct drive
    echo - letter.
    echo -------------------------------------------------------------------------------
    echo.

    REM Save the volume GUID for the drive that has the WinPE boot media. Then
    REM unmount it in case one of the disks to be restored needs that drive letter.
    for /f %%a in ('mountvol %IBM_SP_CLIENT_DRIVE% /l') do set IBM_SP_CLIENT_DRIVE_GUID=%%a
    if "%IBM_SP_DEBUG%" equ "true" (
        echo DEBUG: Dumping environment variables
        set
    )
    mountvol %IBM_SP_CLIENT_DRIVE% /d

    REM Use mountvol to mount local volumes.
    REM Parse tsmvolmap.txt:
    REM     for /f "tokens=1,2,3 delims=|" %%a in (tsmvolmap.txt) do @echo %%a %%b %%c
    REM %%a is the drive letter, %%b is the volume GUID, and %%c is the
    REM "0" or "1" critical volume flag. A drive letter of "\?" is a
    REM hidden partition and should not be mounted.
    REM Then do this:
    REM     MOUNTVOL %%a /d
    REM     MOUNTVOL %%a %%b
    REM The dismount is needed in case the wrong volume is mounted on that drive
    REM letter, in which case the subsequent mount does not work.
    if not exist "%IBM_SP_TSMVOLMAP_FILE%" goto :TSMVOLMAP_TXT_NOT_FOUND

    for /f "tokens=1,2,3 delims=|" %%a in (%IBM_SP_TSMVOLMAP_FILE%) do (
        if "%%a" neq "\?" (
            REM The dismount can "fail" if no volume is actually mounted, but
            REM that is OK. Therefore the output from the first mountvol is
            REM discarded.
            mountvol %%a /d >nul 2>&1
            mountvol %%a %%b
            if exist %%a (
                echo Volume %%b mounted on drive %%a
            ) else (
                echo Could not mount %%b on drive %%a. Check the system and volume map file "%IBM_SP_TSMVOLMAP_FILE%".
            )
        )
    )

    REM Find an available drive letter, and re-mount the winPE media drive,
    REM then update DSM_DIR and PATH accordingly.
    if "%IBM_SP_DEBUG%" equ "true" (
        echo DEBUG: Dumping environment variables again
        set
    )

    set IBM_SP_CLIENT_DRIVE=
    for %%d in (A B C D E F G H I J K L M N O P Q R S T U V W X Y Z) do (
        if not exist "%%d:" set IBM_SP_CLIENT_DRIVE=%%d:
    )
    if "%IBM_SP_DEBUG%" equ "true" (
        echo DEBUG: Looking for an available drive to re-mount the WinPE boot media
    )
    if "%IBM_SP_CLIENT_DRIVE%" equ "" goto :CANNOT_REMOUNT_CD_DRIVE
    mountvol %IBM_SP_CLIENT_DRIVE% %IBM_SP_CLIENT_DRIVE_GUID%
    call :SET_OTHER_ENVIRONMENT_VARIABLES

    echo.


:DO_BUILD_DRIVE_RESTORE_COMMANDS
    set IBM_SP_DRIVE_RESTORE_MACRO=%IBM_SP_RECOVERY_DIR%\dsmc_restore_drives.cmd
    if exist "%IBM_SP_DRIVE_RESTORE_MACRO%" del "%IBM_SP_DRIVE_RESTORE_MACRO%"

    REM The file spaces can include system state, drive letters, network shares,
    REM API file spaces, etc. Look for the file space names that match the
    REM pattern "\\xxxxxxx\y$" where "xxxxxxx" is the Windows host name and "y"
    REM is a drive letter.
    REM
    REM    1     02/01/2019 11:12:19     VSS     XXXXXXX\SystemState\NULL\System State\SystemState
    REM    2     02/01/2019 11:12:19     NTFS    \\xxxxxxx\c$
    REM    3     01/03/2022 19:12:39     NTFS    \\xxxxxxx\d$ 
    REM    4     00/00/0000 00:00:00     API:Sample  apifs
    REM
    REM Or when replication is used:
    REM
    REM    1     01/03/2022 19:09:48     VSS     Current           XXXXXXX\SystemState\NULL\System State\SystemState
    REM    2     01/03/2022 19:12:37     NTFS    Current           \\xxxxxxx\c$ 
    REM    3     01/03/2022 19:12:39     NTFS    Not Current       \\xxxxxxx\d$ 
    REM
    REM The code should match file spaces 2 and 3 in the examples above.
    REM
    REM Version 2.3 fix: The number of tokens in the QUERY FILESPACE output can
    REM vary depending on whether failover mode is in effect and the value of
    REM the replication status for the file space ("Current" or "Not Current",
    REM for example).
    REM
    REM The Microsoft Windows batch file language has no provision to specify
    REM the last non-empty token in a line. Therefore, this code parses the
    REM input line into 10 variables (%%a, %%b, %%c, ..., %%j), which is more
    REM than enough for all cases, then searches for the last non-empty string
    REM that looks like a drive file space name.
    for /f "usebackq tokens=1-10" %%a in (%IBM_SP_FILE_SPACES%) do (
        REM Step 1. Determine if the current input line is for a file space.
        REM That is, if the first token is an integer (1, 2, 3, ...) then the
        REM line is for file space information. Otherwise the input line is
        REM other information from the QUERY FILESPACE command, and can be
        REM ignored.
        REM
        REM The "set /a" command treats the first token, %%a, as an integer and
        REM assigns the value to IBM_SP_TOKEN1. If %%a is not an integer, then
        REM the value 0 (zero) is assigned. Thus if IBM_SP_TOKEN1 is nonzero, it
        REM it must be a file space line. Output is redirected to nul to avoid
        REM unnecessary screen echo of the set value.
        set /a IBM_SP_TOKEN1=%%a > nul 2>&1
        if !IBM_SP_TOKEN1! gtr 0 (
            REM Step 2. The first possible variable that has the file space name
            REM is %%e. In this case, %%f is an empty string. The code checks if
            REM %%f is empty, and if so, assigned %%e to IBM_SP_TEMP. Otherwise
            REM the code checks whether %%g is empty, and so on, until it finds
            REM the last non-empty string.
            set IBM_SP_TEMP=
            if "%%f" equ "" (
                set IBM_SP_TEMP=%%e
            ) else (
                if "%%g" equ "" (
                    set IBM_SP_TEMP=%%f
                ) else (
                    if "%%h" equ "" (
                        set IBM_SP_TEMP=%%g
                    ) else (
                        if "%%i" equ "" (
                            set IBM_SP_TEMP=%%h
                        ) else (
                            if "%%j" equ "" (
                                set IBM_SP_TEMP=%%i
                            ) else (
                                set IBM_SP_TEMP=%%j
                            )
                        )
                    )
                )
            )

            REM Step 3. Check if the first two characters are "\\", the third-to-
            REM lastcharacter is "\", and the last character is "$".
            if "!IBM_SP_TEMP:~0,2!" equ "\\" (
                if "!IBM_SP_TEMP:~-3,1!" equ "\" (
                    if "!IBM_SP_TEMP:~-1!" equ "$" (
                        REM Step 4. If it passes the above tests, assume it is a
                        REM file space for a drive letter, and create the restore
                        REM command.
                        set IBM_SP_TARGET_DRIVE=!IBM_SP_TEMP:~-2,1!:

                        REM Step 5. Cross-check with the tsmvolmap.txt file, to see
                        REM if the drive is in the file, and if it is required.
                        set IBM_SP_COMMENT1=REM ATTENTION - The following drive is not in the volume map file "%IBM_SP_TSMVOLMAP_FILE%"
                        set IBM_SP_COMMENT2=REM ATTENTION - The following drive is not mounted

                        for /f "tokens=1,2,3 delims=|" %%v in (%IBM_SP_TSMVOLMAP_FILE%) do (
                            if /i "%%v" equ "!IBM_SP_TARGET_DRIVE!" (
                                if "%%x" equ "1" (
                                    set IBM_SP_COMMENT1=REM The following drive is required to restore the system
                                ) else (
                                    set IBM_SP_COMMENT1=REM Restore of the following drive is optional
                                )

                                if exist "!IBM_SP_TARGET_DRIVE!" (
                                    set IBM_SP_COMMENT2=REM The drive is mounted.
                                )
                            )
                        )

                        echo !IBM_SP_COMMENT1!>> "%IBM_SP_DRIVE_RESTORE_MACRO%"
                        echo !IBM_SP_COMMENT2!>> "%IBM_SP_DRIVE_RESTORE_MACRO%"
                        echo @echo - Starting restore of drive !IBM_SP_TEMP! ...>> "%IBM_SP_DRIVE_RESTORE_MACRO%"
                        echo dsmc restore !IBM_SP_TEMP!\ !IBM_SP_TEMP:~-2,1!:\ -subdir=yes -replace=no %IBM_SP_PIT_OPTIONS% ^> %IBM_SP_RECOVERY_DIR%\dsmc_restore_drive_!IBM_SP_TEMP:~-2,1!.log 2^>^&^1>> "%IBM_SP_DRIVE_RESTORE_MACRO%"
                        REM ERRORLEVEL must be upper case to avoid ambiguity between variable %%e and %%errorlevel
                        echo @echo - Restore of drive !IBM_SP_TEMP! finished with return code %%ERRORLEVEL%%>> "%IBM_SP_DRIVE_RESTORE_MACRO%"
                        echo.>> "%IBM_SP_DRIVE_RESTORE_MACRO%"
                    )
                )
            )
        )
    )


:DO_DRIVE_RESTORES
    call :DISPLAY_STEP_NUMBER
    echo -------------------------------------------------------------------------------
    echo - A script named "%IBM_SP_DRIVE_RESTORE_MACRO%" was created to
    echo - restore the local drives. Review the script for accuracy. You can also make
    echo - changes, if necessary. When you are finished, save any changes, then close
    echo - Notepad to continue.
    echo -
    echo - Tip: The output from each drive restore operation is redirected to a log
    echo - file. If the drive has a very large number of files, such as one million or
    echo - more files, then add the -QUIET option to the restore command for that
    echo - drive. -QUIET suppresses the logging of each restored file, thus preventing
    echo - the output file from filling the X: drive.
    echo -------------------------------------------------------------------------------
    echo.
    set /p IBM_SP_USER_RESP=Press 'ENTER' to review the drive restore script =^> 
    echo.

    REM Display the script for the user for review.
    notepad "%IBM_SP_DRIVE_RESTORE_MACRO%"

    echo.
    set /p IBM_SP_USER_RESP=Press 'ENTER' to start the drive restore =^> 
    echo.

    call "%IBM_SP_DRIVE_RESTORE_MACRO%"


:COPY_BOOTMGR_FILE
    REM If there is a hidden system partition, copy the appropriate boot manager
    REM file to that partition. Use the info from tsmvolmap.txt to figure this
    REM out. In tsmvolmap.txt, the hidden partition has a drive letter of "\?"
    REM and a critical flag "1". At this point in the restore, the volume will
    REM have a directory named "Boot" in its root.
    REM
    REM Note that there is no definitive way to identify the system partition,
    REM but the above conditions are a sufficient "good guess".
    if "%IBM_SP_FIRMWARE%" neq "BIOS" goto :COPY_UEFI_BOOTMGR_FILE

    set IBM_SP_BOOTMGR_FILE=C:\Windows\Boot\PCAT\bootmgr

    for /f "tokens=1,2,3 delims=|" %%a in (%IBM_SP_TSMVOLMAP_FILE%) do (
        if "%%a" equ "\?" (
            if "%%c" equ "1" (
                if exist %%b\Boot\* (
                    if not exist %%b\Boot\bootmgr (
                        if exist %IBM_SP_BOOTMGR_FILE% (
                            copy %IBM_SP_BOOTMGR_FILE% %%b >nul 2>&1
                            if exist %%b\bootmgr (
                                echo Successfully copied "%IBM_SP_BOOTMGR_FILE%" to volume "%%b".
                            ) else (
                                echo ERROR: Could not copy "%IBM_SP_BOOTMGR_FILE%" to volume "%%b". This is unexpected.
                            )
                        ) else (
                            echo WARNING: Could not find "%IBM_SP_BOOTMGR_FILE%" to copy to "%%bBoot". This is unexpected.
                        )
                    )
                )
            )
        )
    )

    goto :EPILOG

:COPY_UEFI_BOOTMGR_FILE
    if "%IBM_SP_FIRMWARE%" neq "UEFI" goto :EPILOG
    if "%IBM_SP_EFI_DRIVE%" equ "" call :CANNOT_CHECK_EFI_PARTITION


:EPILOG
    echo.
    echo -------------------------------------------------------------------------------
    echo - The restore is finished. The restore output logs are here:
    echo -
    for %%f in (%IBM_SP_RECOVERY_DIR%\dsmc_restore_*.log) do @echo -    %%f
    echo -    %DSM_LOG%\dsmerror.log
    echo -
    echo - Review the output logs to verify the success of the restore. When you are
    echo - satisfied that the restore is complete, restart the system.
    echo -------------------------------------------------------------------------------
    echo.
    goto :DONE


:DISPLAY_STEP_NUMBER
    set /a IBM_SP_STEP_NUMBER+=1
    echo.
    echo [1m[104m                                                                               [0m
    echo [1m[93m    [7m Step %IBM_SP_STEP_NUMBER% [7m[0m
    echo [1m[104m                                                                               [0m
    echo.
    goto :EOF


:NOT_RUNNING_IN_WINPE_MODE
    echo.
    echo -------------------------------------------------------------------------------
    echo - The system is not running in WinPE. This script runs only when the system is
    echo - booted into WinPE.
    echo -------------------------------------------------------------------------------
    echo.
    REM Go to EOF instead of DONE since there is no need to create a debug bundle.
    goto :EOF


:FIRMWARE_UNKNOWN
    echo.
    echo -------------------------------------------------------------------------------
    echo - Ask the administrator of the system you are restoring for the system's
    echo - firmware type. Make sure you boot WinPE with the correct firmware type, then
    echo - run this script again.
    echo -------------------------------------------------------------------------------
    echo.
    goto :DONE


:USER_QUIT
    echo.
    echo -------------------------------------------------------------------------------
    echo - Exiting the restore script.
    echo -------------------------------------------------------------------------------
    echo.
    goto :DONE


:CANNOT_FIND_CD_DRIVE
    echo.
    echo -------------------------------------------------------------------------------
    echo - A drive letter containing the IBM Spectrum Protect Backup-Archive Client
    echo - could not be found. Make sure the drive letter is available, and has the
    echo - path "Program Files\Tivoli\TSM\baclient\dsmc.exe" in it. Then run this script
    echo - again.
    echo -------------------------------------------------------------------------------
    echo.
    goto :DONE


:CANNOT_CREATE_WORKING_DIR
    echo.
    echo -------------------------------------------------------------------------------
    echo - Unable to create the working directory "%IBM_SP_RECOVERY_DIR%". The restore
    echo - cannot proceed. Identify the cause of the problem, then run this script
    echo - again.
    echo -------------------------------------------------------------------------------
    echo.
    goto :DONE


:CANNOT_CONNECT_TO_PROTECT_SERVER
    echo.
    echo -------------------------------------------------------------------------------
    echo - The %IBM_SP_NAME% server cannot be reached, so the script cannot continue.
    echo - After you set up the network so that the server can be reached, you can run
    echo - this script again.
    echo -------------------------------------------------------------------------------
    echo.
    goto :DONE


:CANNOT_QUERY_SESSION
    echo.
    echo -------------------------------------------------------------------------------
    echo - The %IBM_SP_NAME% backup-archive client could not connect to the
    echo - server. Afer you have corrected the problem, you can run this script again.
    echo -------------------------------------------------------------------------------
    echo.
    goto :DONE


:CANNOT_REMOUNT_CD_DRIVE
    echo.
    echo -------------------------------------------------------------------------------
    echo - The WinPE boot media drive needs to be remounted, but no drive letter is
    echo - available. Afer you have corrected the problem, you can run this script
    echo - again.
    echo -------------------------------------------------------------------------------
    echo.
    goto :DONE



:SYSTEM_STATE_RESTORE_FAILED
    echo.
    echo -------------------------------------------------------------------------------
    echo - The system state restore failed. Review the dsmerror.log file. After you have
    echo - corrected the problem, you can run this script again. You might need to first
    echo - reboot WinPE.
    echo -------------------------------------------------------------------------------
    echo.
    goto :DONE


:TSMVOLMAP_TXT_NOT_FOUND
    echo.
    echo -------------------------------------------------------------------------------
    echo - The tsmvolmap.txt file could not be found. It should have been restored to
    echo - a subdirectory of '%TEMP%'
    echo - and then copied to '%IBM_SP_TSMVOLMAP_FILE%'.
    echo -
    echo - Without that file, this script cannot continue. Instead, you will need to:
    echo -
    echo -    1. Manually mount the local drives
    echo -    2. Restore the local drives
    echo -    3. Copy the bootmgr file to the hidden partition (if applicable)
    echo -------------------------------------------------------------------------------
    echo.
    goto :DONE


:CANNOT_CHECK_EFI_PARTITION
    echo.
    echo -------------------------------------------------------------------------------
    echo - No empty drive letter is available to mount the EFI system partition. You can
    echo - manually check whether EFI files need to be restored by following these
    echo - steps:
    echo -
    echo -    1. Run the command MOUNTVOL to see which volume GUIDs are mounted on each
    echo -       drive letter.
    echo -
    echo -    2. Choose a drive letter and note the volume GUID mounted on it.
    echo -
    echo -    3. Run the command:
    echo -
    echo -           MOUNTVOL X: /D
    echo -
    echo -       to dismount the file system from a drive letter. X: is the drive letter
    echo -       selected in step 2.
    echo -
    echo -    4. Run the command:
    echo -
    echo -           MOUNTVOL X: /S
    echo -
    echo -    5. Run the command:
    echo -
    echo -           dsmc restore "{%IBM_SP_SYSSTATE_FILESPACE%}\EFI\Microsoft\*" X:\EFI\ -subdir=yes -replace=no %IBM_SP_PIT_OPTIONS%
    echo -
    echo -    6. Run these commands:
    echo -
    echo -           MOUNTVOL X: /D
    echo -           MOUNTVOL X: volume_guid
    echo -
    echo -       where "volume_guid" is the original volume GUID from step 2.
    echo -------------------------------------------------------------------------------
    echo.
    set /p IBM_SP_USER_RESP=Press 'ENTER' to continue =^> 
    echo.
    goto :EOF


:FIND_CD_DRIVE
    set IBM_SP_CLIENT_DRIVE=
    for %%d in (A B C D E F G H I J K L M N O P Q R S T U V W X Y Z) do (
        if exist "%%d:\%IBM_SP_DUMMY_ID_FILE%" (
            set IBM_SP_CLIENT_DRIVE=%%d:
        )
    )

    goto :EOF


:SET_SYSTEM_STATE_FILE_SPACE_NAME
    REM The file spaces can include system state, drive letters, network shares,
    REM API file spaces, etc. Look for the file space name that matches the
    REM pattern "xxxxxxx\SystemState\NULL\System State\SystemState" where
    REM "xxxxxxx" is the Windows host name.
    REM
    REM    1     02/01/2019 11:12:19     VSS     XXXXXXX\SystemState\NULL\System State\SystemState
    REM    2     02/01/2019 11:12:19     NTFS    \\xxxxxxx\c$
    REM    3     01/03/2022 19:12:39     NTFS    \\xxxxxxx\d$ 
    REM    4     00/00/0000 00:00:00     API:Sample  apifs
    REM
    REM Or when replication is used:
    REM
    REM    1     01/03/2022 19:09:48     VSS     Current           XXXXXXX\SystemState\NULL\System State\SystemState
    REM    2     01/03/2022 19:12:37     NTFS    Current           \\xxxxxxx\c$ 
    REM    3     01/03/2022 19:12:39     NTFS    Not Current       \\xxxxxxx\d$ 
    REM
    REM The code should match file space 1 in the examples above.
    REM
    REM Note: the code assumes that there is exactly one system state file space
    REM for the node.
    REM
    REM Version 2.3 fix: The number of tokens in the QUERY FILESPACE output can
    REM vary depending on whether failover mode is in effect and the value of
    REM the replication status for the file space ("Current" or "Not Current",
    REM for example).
    REM
    REM The Microsoft Windows batch file language has no provision to specify
    REM the last non-empty token in a line. Therefore, this code parses the
    REM input line into 10 variables (%%a, %%b, %%c, ..., %%j), which is more
    REM than enough for all cases, then searches for the file space name.
    for /f "usebackq tokens=1-10" %%a in (%IBM_SP_FILE_SPACES%) do (
        REM Step 1. Determine if the current input line is for a file space.
        REM That is, if the first token is an integer (1, 2, 3, ...) then the
        REM line is for file space information. Otherwise the input line is
        REM other information from the QUERY FILESPACE command, and can be
        REM ignored.
        REM
        REM The "set /a" command treats the first token, %%a, as an integer and
        REM assigns the value to IBM_SP_TOKEN1. If %%a is not an integer, then
        REM the value 0 (zero) is assigned. Thus if IBM_SP_TOKEN1 is nonzero, it
        REM it must be a file space line. Output is redirected to nul to avoid
        REM unnecessary screen echo of the set value.
        set /a IBM_SP_TOKEN1=%%a > nul 2>&1
        if !IBM_SP_TOKEN1! gtr 0 (
            REM STEP 2. Check if this is a VSS file space.
            if "%%d" equ "VSS" (
                REM Step 3. The system state file space name is parsed into two
                REM parts: "xxxxxxx\SystemState\NULL\System" and
                REM "State\SystemState".
                REM
                REM Starting with %%f, check if the value is 
                REM "State\SystemState". If yes, then variables %%e and %%f are
                REM combined to form the full file space name. If replication
                REM status is included in the output, then this check does not
                REM pass, in which case %%g is checked, then %%h, and so on.
                set IBM_SP_SYSSTATE_FILESPACE=
                if "%%f" equ "State\SystemState" (
                    set IBM_SP_SYSSTATE_FILESPACE=%%e %%f
                ) else (
                    if "%%g" equ "State\SystemState" (
                        set IBM_SP_SYSSTATE_FILESPACE=%%f %%g
                    ) else (
                        if "%%h" equ "State\SystemState" (
                            set IBM_SP_SYSSTATE_FILESPACE=%%g %%h
                        ) else (
                            if "%%i" equ "State\SystemState" (
                                set IBM_SP_SYSSTATE_FILESPACE=%%h %%i
                            ) else (
                                if "%%j" equ "State\SystemState" (
                                    set IBM_SP_SYSSTATE_FILESPACE=%%i %%j
                                )
                            )
                        )
                    )
                )
            )
        )
    )

    goto :EOF


:SET_OTHER_ENVIRONMENT_VARIABLES
    if exist "%IBM_SP_CLIENT_DRIVE%\Program Files 64\Tivoli\TSM\baclient\dsmc.exe" (
        set DSM_DIR=%IBM_SP_CLIENT_DRIVE%\Program Files 64\Tivoli\TSM\baclient
    ) else (
        set DSM_DIR=%IBM_SP_CLIENT_DRIVE%\Program Files\Tivoli\TSM\baclient
    )

    set PATH=%IBM_SP_ORIGINAL_PATH%;%DSM_DIR%
    goto :EOF


:DONE
    REM Create a .zip file archive of the recovery directory.
    if exist "%IBM_SP_RECOVERY_DIR%" (
        powershell Compress-Archive -Path "%IBM_SP_RECOVERY_DIR%" -DestinationPath "%IBM_SP_ZIP_BUNDLE%"
        if exist "%IBM_SP_ZIP_BUNDLE%" (
            echo Archive bundle "%IBM_SP_ZIP_BUNDLE%" was created for directory "%IBM_SP_RECOVERY_DIR%".
        ) else (
            echo ERROR: Could not create the .zip file archive of "%IBM_SP_RECOVERY_DIR%".
        )
    )
    goto :EOF